/* crypto/engine/hw_etrax_fs.c */
/* ====================================================================
 * Copyright (c) 2003 Axis Communications AB.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit. (http://www.OpenSSL.org/)"
 *
 * 4. The names "OpenSSL Toolkit" and "OpenSSL Project" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    licensing@OpenSSL.org.
 *
 * 5. Products derived from this software may not be called "OpenSSL"
 *    nor may "OpenSSL" appear in their names without prior written
 *    permission of the OpenSSL Project.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the OpenSSL Project
 *    for use in the OpenSSL Toolkit (http://www.OpenSSL.org/)"
 *
 * THIS SOFTWARE IS PROVIDED BY THE OpenSSL PROJECT ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE OpenSSL PROJECT OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).  This product includes software written by Tim
 * Hudson (tjh@cryptsoft.com).
 *
 */

#include <openssl/objects.h>
#include <openssl/engine.h>
#include <openssl/evp.h>

#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifndef OPENSSL_NO_HW
#ifndef OPENSSL_NO_HW_ETRAX_FS

#include <sys/ioctl.h>
#include <asm/arch/cryptocop.h>


struct efs_cipher
	{
	unsigned int	id;
	int		nid;
	int		keylength;
	int		ivlength;
	};

struct efs_digest
	{
	unsigned int	id;
	int		nid;
	};

struct efseng_state
	{
	cryptocop_session_id sid;
	int fd;
	};

static int efseng_init_key(EVP_CIPHER_CTX *ctx, const unsigned char *key,
	const unsigned char *iv, int enc);
static int efseng_cipher(EVP_CIPHER_CTX *ctx, unsigned char *out,
	const unsigned char *in, unsigned int inl);
static int efseng_cleanup(EVP_CIPHER_CTX *ctx);

	
static const char efs_engine_id[] = "etrax_fs";
static const char efs_engine_name[] = "AXIS ETRAX FS stream co-processor engine";

#define EFSENG_CIPHER_ID(alg,mode,kl) (alg | (mode << 4) | (kl << 8))
#define ALG_FROM_CIPHER_ID(cid) (cid & 0x0003)
#define MODE_FROM_CIPHER_ID(cid) ((cid >> 4) & 0x0003)
#define KL_FROM_CIPHER_ID(cid) ((cid >> 8) & 0xffff)

static const ENGINE_CMD_DEFN efseng_defns[] = {
	{ 0, NULL, NULL, 0 }
};

static const struct efs_cipher ciphers[] =
	{
	{
	EFSENG_CIPHER_ID(cryptocop_cipher_des, cryptocop_cipher_mode_ecb, 64),
	NID_des_ecb,
	8,
	0
	},
	{
	EFSENG_CIPHER_ID(cryptocop_cipher_des, cryptocop_cipher_mode_cbc, 64),
	NID_des_cbc,
	8,
	8
	},
	{
	EFSENG_CIPHER_ID(cryptocop_cipher_3des, cryptocop_cipher_mode_ecb,192),
	NID_des_ede3_ecb,
	24,
	0
	},
	{
	EFSENG_CIPHER_ID(cryptocop_cipher_3des, cryptocop_cipher_mode_cbc, 192),
	NID_des_ede3_cbc,
	24,
	8
	},
	{
	EFSENG_CIPHER_ID(cryptocop_cipher_aes, cryptocop_cipher_mode_ecb, 128),
	NID_aes_128_ecb,
	16,
	0
	},
	{
	EFSENG_CIPHER_ID(cryptocop_cipher_aes, cryptocop_cipher_mode_cbc, 128),
	NID_aes_128_cbc,
	16,
	16
	},
	{
	EFSENG_CIPHER_ID(cryptocop_cipher_aes, cryptocop_cipher_mode_ecb, 192),
	NID_aes_192_ecb,
	24,
	0
	},
	{
	EFSENG_CIPHER_ID(cryptocop_cipher_aes, cryptocop_cipher_mode_cbc, 192),
	NID_aes_192_cbc,
	24,
	16
	},
	{
	EFSENG_CIPHER_ID(cryptocop_cipher_aes, cryptocop_cipher_mode_ecb, 256),
	NID_aes_256_ecb,
	32,
	0
	},
	{
	EFSENG_CIPHER_ID(cryptocop_cipher_aes, cryptocop_cipher_mode_cbc, 256),
	NID_aes_256_cbc,
	32,
	16
	}
	};

static const struct efs_digest digests[] =
	{
	{
	cryptocop_digest_md5,
	NID_md5
	},
	{
	cryptocop_digest_sha1,
	NID_sha1
	}
	};

/* DES ECB EVP */
const EVP_CIPHER efseng_des_ecb =
	{
	NID_des_ecb,
	8, 8, 0,
	EVP_CIPH_ECB_MODE,
	efseng_init_key,
	efseng_cipher,
	efseng_cleanup,
	sizeof(struct efseng_state),
	EVP_CIPHER_set_asn1_iv,
	EVP_CIPHER_get_asn1_iv,
	NULL
};

/* DES CBC EVP */
const EVP_CIPHER efseng_des_cbc =
	{
	NID_des_cbc,
	8, 8, 8,
	EVP_CIPH_CBC_MODE,
	efseng_init_key,
	efseng_cipher,
	efseng_cleanup,
	sizeof(struct efseng_state),
	EVP_CIPHER_set_asn1_iv,
	EVP_CIPHER_get_asn1_iv,
	NULL
};

/* 3DES ECB EVP */
const EVP_CIPHER efseng_3des_ecb =
	{
	NID_des_ede3_ecb,
	8, 24, 0,
	EVP_CIPH_ECB_MODE,
	efseng_init_key,
	efseng_cipher,
	efseng_cleanup,
	sizeof(struct efseng_state),
	EVP_CIPHER_set_asn1_iv,
	EVP_CIPHER_get_asn1_iv,
	NULL
};

/* 3DES CBC EVP */
const EVP_CIPHER efseng_3des_cbc =
	{
	NID_des_ede3_cbc,
	8, 24, 8,
	EVP_CIPH_CBC_MODE,
	efseng_init_key,
	efseng_cipher,
	efseng_cleanup,
	sizeof(struct efseng_state),
	EVP_CIPHER_set_asn1_iv,
	EVP_CIPHER_get_asn1_iv,
	NULL
};

/* AES-128 ECB EVP */
const EVP_CIPHER efseng_aes128_ecb =
	{
	NID_aes_128_ecb,
	16, 16, 0,
	EVP_CIPH_ECB_MODE,
	efseng_init_key,
	efseng_cipher,
	efseng_cleanup,
	sizeof(struct efseng_state),
	EVP_CIPHER_set_asn1_iv,
	EVP_CIPHER_get_asn1_iv,
	NULL
};

/* AES-128 CBC EVP */
const EVP_CIPHER efseng_aes128_cbc =
	{
	NID_aes_128_cbc,
	16, 16, 16,
	EVP_CIPH_CBC_MODE,
	efseng_init_key,
	efseng_cipher,
	efseng_cleanup,
	sizeof(struct efseng_state),
	EVP_CIPHER_set_asn1_iv,
	EVP_CIPHER_get_asn1_iv,
	NULL
};

/* AES-192 ECB EVP */
const EVP_CIPHER efseng_aes192_ecb =
	{
	NID_aes_192_ecb,
	16, 24, 0,
	EVP_CIPH_ECB_MODE,
	efseng_init_key,
	efseng_cipher,
	efseng_cleanup,
	sizeof(struct efseng_state),
	EVP_CIPHER_set_asn1_iv,
	EVP_CIPHER_get_asn1_iv,
	NULL
};

/* AES-192 CBC EVP */
const EVP_CIPHER efseng_aes192_cbc =
	{
	NID_aes_192_cbc,
	16, 24, 16,
	EVP_CIPH_CBC_MODE,
	efseng_init_key,
	efseng_cipher,
	efseng_cleanup,
	sizeof(struct efseng_state),
	EVP_CIPHER_set_asn1_iv,
	EVP_CIPHER_get_asn1_iv,
	NULL
};

/* AES-256 ECB EVP */
const EVP_CIPHER efseng_aes256_ecb =
	{
	NID_aes_256_ecb,
	16, 32, 0,
	EVP_CIPH_ECB_MODE,
	efseng_init_key,
	efseng_cipher,
	efseng_cleanup,
	sizeof(struct efseng_state),
	EVP_CIPHER_set_asn1_iv,
	EVP_CIPHER_get_asn1_iv,
	NULL
};

/* AES-256 CBC EVP */
const EVP_CIPHER efseng_aes256_cbc =
	{
	NID_aes_256_cbc,
	16, 32, 16,
	EVP_CIPH_CBC_MODE,
	efseng_init_key,
	efseng_cipher,
	efseng_cleanup,
	sizeof(struct efseng_state),
	EVP_CIPHER_set_asn1_iv,
	EVP_CIPHER_get_asn1_iv,
	NULL
};

#if 0
/* MD5 EVP */
const EVP_MD efseng_md5 =
	{
	NID_md5,
};
#endif

static const int efseng_cipher_nids[] =
	{
	NID_des_ecb,
	NID_des_cbc,
	NID_des_ede3_ecb,
	NID_des_ede3_cbc,
	NID_aes_128_ecb,
	NID_aes_128_cbc,
	NID_aes_192_ecb,
	NID_aes_192_cbc,
	NID_aes_256_ecb,
	NID_aes_256_cbc
	};

static const int efseng_digest_nids[] =
	{
#if 0
	NID_md5,
	NID_sha1
#else
	NID_undef
#endif
	};

static int efseng_max_iv(unsigned int cipher_id)
	{
	int i;
	for (i = 0; i < sizeof(ciphers)/sizeof(*ciphers); i++)
		{
		if (ciphers[i].id == cipher_id)
			{
			return ciphers[i].ivlength;
			}
		}
	return (0);
	}

static unsigned int nid_to_cipher_id(int nid)
	{
	int i;
	for (i = 0; i<sizeof(ciphers)/sizeof(struct efs_cipher); i++)
		{
		if (ciphers[i].nid == nid)
			{
			return (ciphers[i].id);
			}
		}
	return (NID_undef);
	}

static int efseng_usable_ciphers(const int **nids)
	{
	*nids = efseng_cipher_nids;
	return sizeof(efseng_cipher_nids)/sizeof(*efseng_cipher_nids);
	}


static int efseng_ciphers(ENGINE *e, const EVP_CIPHER **cipher,
	const int **nids, int nid)
	{
	if(!cipher)
		return (efseng_usable_ciphers(nids));

	switch (nid){
	case NID_des_ecb:
		*cipher = &efseng_des_ecb;
		break;
	case NID_des_cbc:
		*cipher = &efseng_des_cbc;
		break;
	case NID_des_ede3_ecb:
		*cipher = &efseng_3des_ecb;
		break;
	case NID_des_ede3_cbc:
		*cipher = &efseng_3des_cbc;
		break;
	case NID_aes_128_ecb:
		*cipher = &efseng_aes128_ecb;
		break;
	case NID_aes_128_cbc:
		*cipher = &efseng_aes128_cbc;
		break;
	case NID_aes_192_ecb:
		*cipher = &efseng_aes192_ecb;
		break;
	case NID_aes_192_cbc:
		*cipher = &efseng_aes192_cbc;
		break;
	case NID_aes_256_ecb:
		*cipher = &efseng_aes256_ecb;
		break;
	case NID_aes_256_cbc:
		*cipher = &efseng_aes256_cbc;
		break;
	default:
		*cipher = NULL;
		break;
	}
	return (*cipher != NULL);
	}


static int efseng_usable_digests(const int **nids)
	{
	*nids = efseng_digest_nids;
	return sizeof(efseng_digest_nids)/sizeof(*efseng_digest_nids);
	}


static int
efseng_digests(ENGINE *e, const EVP_MD **digest,
    const int **nids, int nid)
	{
	if (!digest)
		return (efseng_usable_digests(nids));

	switch (nid) {
	case NID_md5:
//		*digest = &efseng_md5;
		*digest = NULL;
		break;
	case NID_sha1:
//		*digest = &efseng_sha1;
		*digest = NULL;		
		 break;
	default:
		*digest = NULL;
		break;
	}
	return (*digest != NULL);
}


static int efseng_open_dev(void)
	{
	return open("/dev/cryptocop", O_RDONLY);
	}

static int efseng_init_key(EVP_CIPHER_CTX *ctx, const unsigned char *key,
			   const unsigned char *iv, int enc)
	{
	int s;
	struct strcop_session_op sop =
		{
		des3_mode: cryptocop_3des_ede,
		csum: cryptocop_csum_none,
		key: NULL,
		keylen: 0,
		ses_id: 0
		};
	int err;

	struct efseng_state *state = ctx->cipher_data;
	unsigned int cipher_id;

	if ((cipher_id = nid_to_cipher_id(ctx->cipher->nid)) == NID_undef)
		return (0);

	if (ctx->cipher->iv_len < efseng_max_iv(cipher_id))
		return (0);

	if (KL_FROM_CIPHER_ID(cipher_id)/8 != ctx->key_len)
		return (0);

	sop.cipher = ALG_FROM_CIPHER_ID(cipher_id);
	sop.cmode = MODE_FROM_CIPHER_ID(cipher_id);
	sop.key = (unsigned char *)key;
	sop.keylen = KL_FROM_CIPHER_ID(cipher_id);

	s = efseng_open_dev();
	if (s < 0)
		{
		return (0);
		}
	if ((err = ioctl(s,
		_IO(ETRAXCRYPTOCOP_IOCTYPE, CRYPTOCOP_IO_CREATE_SESSION),
		&sop)) == -1)
		{
		return (0);
		}
	state->fd = s;
	state->sid = sop.ses_id;

	return (1);
	}

static int efseng_cipher(EVP_CIPHER_CTX *ctx, unsigned char *out,
	const unsigned char *in, unsigned int inl)
	{
	struct strcop_crypto_op operation = {0};
	int err;
	struct efseng_state *state = ctx->cipher_data;
	void *iiv;
	unsigned char save_iv[EVP_MAX_IV_LENGTH];

	if (state->fd < 0)
		return (0);
	if (!inl)
		return (1);
	if ((inl % ctx->cipher->block_size) != 0)
		return (0);
	if (ctx->cipher->iv_len > CRYPTOCOP_MAX_IV_LENGTH)
		return (0);

	operation.ses_id = state->sid;
	operation.indata = (unsigned char*)in;
	operation.inlen = inl;
	operation.cipher_outdata = out;
	operation.cipher_outlen = inl;
	operation.do_digest = 0;
	operation.do_csum = 0;

	operation.decrypt = ctx->encrypt ? 0 : 1;
	operation.cipher_start = 0;
	operation.cipher_len = inl;
	operation.do_cipher = 1;
	
	if (ctx->cipher->iv_len)
		{
		operation.cipher_explicit = 1;
		memcpy(operation.cipher_iv, ctx->iv, ctx->cipher->iv_len);
		if (!ctx->encrypt)
			{
			iiv = (void *) in + inl - ctx->cipher->iv_len;
			memcpy(save_iv, iiv, ctx->cipher->iv_len);
			}
		}

	if ((err = ioctl(state->fd,
		_IO(ETRAXCRYPTOCOP_IOCTYPE, CRYPTOCOP_IO_PROCESS_OP),
		&operation)) == -1)
		{
		return (0);
		}
	if (ctx->cipher->iv_len)
		{
		if (ctx->encrypt)
			iiv = (void *) out + inl - ctx->cipher->iv_len;
		else
			iiv = save_iv;
		memcpy(ctx->iv, iiv, ctx->cipher->iv_len);
		}
	return (1);
	}

static int efseng_cleanup(EVP_CIPHER_CTX *ctx)
	{
	int ret = 0;
	int err;
	struct efseng_state *state = ctx->cipher_data;
	struct strcop_session_op sop = {0};

	if (state->fd < 0)
		return (0);

	sop.ses_id = state->sid;

	if ((err = ioctl(state->fd,
		_IO(ETRAXCRYPTOCOP_IOCTYPE, CRYPTOCOP_IO_CLOSE_SESSION),
		&sop)) == -1)
		{
		ret = 0;
		}
	else
		{
		ret = 1;
		}	
	close(state->fd);
	state->fd = -1;

	return (ret);
	}



static int efseng_ctrl(ENGINE *e, int cmd, long i, void *p, void (*f)())
	{
	return (1);
	}

/* This internal function is used by engine_etrax_fs() and possibly by the
 * "dynamic" ENGINE support too */
static int bind_helper(ENGINE *e)
	{
	if (!ENGINE_set_id(e, efs_engine_id) ||
		!ENGINE_set_name(e, efs_engine_name) ||
		!ENGINE_set_ciphers(e, efseng_ciphers) ||
		!ENGINE_set_digests(e, efseng_digests) ||
		!ENGINE_set_ctrl_function(e, efseng_ctrl) ||
		!ENGINE_set_cmd_defns(e, efseng_defns)) {
	ENGINE_free(e);
	return (0);
	}
	return (1);
	}

#ifndef ENGINE_DYNAMIC_SUPPORT
static ENGINE *engine_etrax_fs(void)
	{
	ENGINE *ret = ENGINE_new();
	if(!ret)
		return NULL;
	if(!bind_helper(ret))
		{
		ENGINE_free(ret);
		return NULL;
		}
	return ret;
	}

void ENGINE_load_etrax_fs(void)
	{
	/* Copied from hw_nuron.c */
	ENGINE *toadd = engine_etrax_fs();
	if(!toadd) return;
	ENGINE_add(toadd);
	ENGINE_free(toadd);
	ERR_clear_error();
	}
#endif

/* This stuff is needed if this ENGINE is being compiled into a self-contained
 * shared-library. */      
#ifdef ENGINE_DYNAMIC_SUPPORT
static int bind_fn(ENGINE *e, const char *id)
	{
	if(id && (strcmp(id, efs_engine_id) != 0))
		return 0;
	if(!bind_helper(e))
		return 0;
	return 1;
	}
IMPLEMENT_DYNAMIC_CHECK_FN()
IMPLEMENT_DYNAMIC_BIND_FN(bind_fn)
#endif /* ENGINE_DYNAMIC_SUPPORT */

#endif /* OPENSSL_NO_HW_ETRAX_FS */
#endif /* OPENSSL_NO_HW */
